                   (* +--------------------------+
                      |  Expressions boolennes  |
                      +--------------------------+ *)

type binop = Et | Ou | Oubien | Impl | Equiv;;
type formule =
  | Vrai | Faux
  | Var  of string
  | Non  of formule
  | Bin  of binop * formule * formule
;;

                    (* +-----------------------+
                       |  Notation simplifie  |
                       +-----------------------+ *)

let prefix *   f g = Bin(Et,    f,g)
and prefix +   f g = Bin(Ou,    f,g)
and prefix ++  f g = Bin(Oubien,f,g)
and prefix =>  f g = Bin(Impl,  f,g)
and prefix <=> f g = Bin(Equiv, f,g)
;;



                    (* +-----------------------+
                       |  Affiche une formule  |
                       +-----------------------+ *)

let rec affiche par f = match f with
  | Vrai   -> print_string "Vrai"
  | Faux   -> print_string "Faux"
  | Var(x) -> print_string x
  | Non(g) -> print_string "Non("; affiche false g; print_string ")"
  | Bin(op,g,h) ->
      if par then print_string "(";
      affiche true g;
      print_string(match op with
		     | Et     -> " * "
		     | Ou     -> " + "
		     | Oubien -> " ++ "
		     |Impl    -> " => " 
		     |Equiv   -> " <=> ");
      affiche true h;
      if par then print_string ")"
;;
let print_formule f = affiche false f; print_newline();;

                           (* +---------+
                              |  Essai  |
                              +---------+ *)

let p = Var "p"
and q = Var "q"
and r = Var "r";;

let f = (p => q) => (q => p);;
let g = (p => q) <=> (Non(q) => Non(p));;

print_formule f;;
print_formule g;;

                       (* +----------------+
                          |  Substitution  |
                          +----------------+ *)

let rec subs x y f = match f with
  | Var(v) when v = x -> y
  | Non(g) -> Non(subs x y g)
  | Bin(op,g,h) -> Bin(op, subs x y g, subs x y h)
  | _ -> f
;;

                        (* +--------------+
                           |  Evaluation  |
                           +--------------+ *)

let rec evalue f = match f with
  | Vrai   -> true
  | Faux   -> false
  | Var x  -> failwith (x ^ " est inconnu")
  | Non(g) -> not(evalue g)
  | Bin(op,g,h) ->
      let a = evalue g and b = evalue h in
      match op with
	| Et     -> a & b
	| Ou     -> a or b
	| Oubien -> a <> b
	| Impl   -> a <= b
	| Equiv  -> a = b
;;

                         (* +------------+
                            |  Identit  |
                            +------------+ *)

let rec identique f g vars = match vars with
  | [] -> evalue(f) = evalue(g)
  | x::suite ->  identique (subs x Vrai f) (subs x Vrai g) suite
              && identique (subs x Faux f) (subs x Faux g) suite
;;

type Resultat = Identique | Diffrent of bool list;;
let rec compare f g vars = match vars with
  | [] -> if evalue(f) = evalue(g) then Identique else Diffrent[]
  | x::suite ->
      match compare (subs x Vrai f) (subs x Vrai g) suite with
	| Diffrent(l) -> Diffrent(true::l)
	| Identique ->
      match compare (subs x Faux f) (subs x Faux g) suite with
	| Diffrent(l) -> Diffrent(false::l)
	| Identique -> Identique
;;

                        (* +--------------+
                           |  tautologie  |
                           +--------------+ *)

let rec tautologie f vars = match vars with
  | [] -> evalue(f)
  | x::suite -> tautologie (subs x Vrai f) suite
              & tautologie (subs x Faux f) suite
;;


                        (* +--------------+
                           |  Rfutation  |
                           +--------------+ *)

let rec rfute f vars = match vars with
  | [] -> not(evalue(f)),[]
  | x::suite ->
      let t,v = rfute (subs x Vrai f) suite in
      if t then t,(x^"=Vrai")::v
      else let t,v = rfute (subs x Faux f) suite in
      if t then t,(x^"=Faux")::v
      else t,[]
;;

                      (* +------------------+
                         |  Simplification  |
                         +------------------+ *)

let neg f = match f with
  | Vrai   -> Faux
  | Faux   -> Vrai
  | Non(g) -> g
  | _      -> Non(f)
;;

let rec simplifie f = match f with
  | Non(g) -> neg(simplifie g)
  | Bin(op,g,h) -> begin match op,simplifie(g),simplifie(h) with

      | Et, Vrai, h     -> h
      | Et, Faux, h     -> Faux
      | Et, g, Vrai     -> g
      | Et, g, Faux     -> Faux
    
      | Ou, Vrai, h     -> Vrai
      | Ou, Faux, h     -> h
      | Ou, g, Vrai     -> Vrai
      | Ou, g, Faux     -> g

      | Oubien, Faux, h -> h
      | Oubien, Vrai, h -> neg h
      | Oubien, g, Faux -> g
      | Oubien, g, Vrai -> neg g

      | Impl, Faux, h   -> Vrai
      | Impl, Vrai, h   -> h
      | Impl, g, Faux   -> neg g
      | Impl, g, Vrai   -> Vrai
      
      | Equiv, Vrai, h  -> h
      | Equiv, Faux, h  -> neg h
      | Equiv, g, Vrai  -> g
      | Equiv, g, Faux  -> neg g

      | _,g,h -> Bin(op,g,h)
    end
  | _ -> f
;;




